@@ -29,7 +29,6 @@
 using std::wstring;
 
 #define ERRNO WSAGetLastError()
-#define strdup _strdup
 
 using namespace mega;
 
@@ -99,8 +98,7 @@ HANDLE ComunicationsManagerNamedPipes::create_new_namedPipe(int *pipeId)
             if (errno == EMFILE) //TODO: review possible out
             {
                 LOG_verbose << " Trying to reduce number of used files by sending ACK to listeners to discard disconnected ones.";
-                string sack="ack";
-                informStateListeners(sack);
+                ackStateListenersAndRemoveClosed();
             }
             if (attempts !=10)
             {
@@ -174,12 +172,12 @@ int ComunicationsManagerNamedPipes::waitForPetition()
             LOG_fatal << "ERROR on connecting to namedPipe. errno: " << ERRNO;
             sleepMilliSeconds(1000);
             pipeGeneral = INVALID_HANDLE_VALUE;
-            return false;
+            return ERRNO;
         }
 
     }
     petitionready = true;
-    return true;
+    return 0;
 }
 
 void ComunicationsManagerNamedPipes::stopWaiting()
@@ -194,10 +192,10 @@ void ComunicationsManagerNamedPipes::stopWaiting()
     }
 }
 
-void ComunicationsManagerNamedPipes::registerStateListener(CmdPetition *inf)
+CmdPetition* ComunicationsManagerNamedPipes::registerStateListener(std::unique_ptr<CmdPetition> &&inf)
 {
-    LOG_debug << "Registering state listener petition with namedPipe: " << ((CmdPetitionNamedPipes *) inf)->outNamedPipe;
-    ComunicationsManager::registerStateListener(inf);
+    LOG_debug << "Registering state listener petition with namedPipe: " << ((CmdPetitionNamedPipes*) inf.get())->outNamedPipe;
+    return ComunicationsManager::registerStateListener(std::move(inf));
 }
 
 //TODO: implement unregisterStateListener, not 100% necesary, since when a state listener is not accessible it is unregistered (to deal with sudden deaths).
@@ -208,9 +206,9 @@ void ComunicationsManagerNamedPipes::registerStateListener(CmdPetition *inf)
  * @brief returnAndClosePetition
  * I will clean struct and close the namedPipe within
  */
-void ComunicationsManagerNamedPipes::returnAndClosePetition(CmdPetition *inf, OUTSTRINGSTREAM *s, int outCode)
+void ComunicationsManagerNamedPipes::returnAndClosePetition(std::unique_ptr<CmdPetition> inf, OUTSTRINGSTREAM *s, int outCode)
 {
-    HANDLE outNamedPipe = ((CmdPetitionNamedPipes *)inf)->outNamedPipe;
+    HANDLE outNamedPipe = ((CmdPetitionNamedPipes*) inf.get())->outNamedPipe;
 
     LOG_verbose << "Output to write in namedPipe " << *(long*)(&outNamedPipe);
 
@@ -226,11 +224,16 @@ void ComunicationsManagerNamedPipes::returnAndClosePetition(CmdPetition *inf, OU
                 connectsucceeded = true;
                 break;
             }
-            else
+            else if (ERRNO == ERROR_NO_DATA)
             {
-                LOG_fatal << "ERROR on connecting to namedPipe " << outNamedPipe << ". errno: " << ERRNO << ". Attempts: " << attempts;
+                break;
             }
-            sleepMilliSeconds(500);
+            else
+            {    // Problem: this, caused by a client shutdown, entails trying to sendPartialOutput again (via MEGAcmdLogger::formatLogToStream)
+                LOG_err << "ERROR on connecting to namedPipe " << outNamedPipe << ". errno: " << ERRNO << ". Attempts: " << attempts;
+                assert(false);
+            }
+            sleepMilliSeconds(50);
         }
         else
         {
@@ -240,8 +243,7 @@ void ComunicationsManagerNamedPipes::returnAndClosePetition(CmdPetition *inf, OU
 
     if (!connectsucceeded)
     {
-        LOG_fatal << "Return and close: Unable to connect on outnamedPipe " << outNamedPipe << " error: " << ERRNO;
-        delete inf;
+        LOG_err << "Return and close: Unable to connect on outnamedPipe " << outNamedPipe << " error: " << ERRNO;
         return;
     }
 
@@ -261,7 +263,6 @@ void ComunicationsManagerNamedPipes::returnAndClosePetition(CmdPetition *inf, OU
         LOG_err << "ERROR writing to namedPipe: " << ERRNO;
     }
     DisconnectNamedPipe(outNamedPipe);
-    delete inf;
 }
 
 
@@ -286,11 +287,16 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, OUTSTRI
                 connectsucceeded = true;
                 break;
             }
+            else if (ERRNO == ERROR_NO_DATA)
+            {
+                break;
+            }
             else
             {
                 cerr << "ERROR on connecting to namedPipe " << outNamedPipe << ". errno: " << ERRNO << ". Attempts: " << attempts << endl;
+                assert(false);
             }
-            sleepMilliSeconds(500);
+            sleepMilliSeconds(50);
         }
         else
         {
@@ -301,20 +307,7 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, OUTSTRI
     if (!connectsucceeded)
     {
         cerr << "sendPartialOutput: Unable to connect on outnamedPipe " << outNamedPipe << " error: " << ERRNO << endl;
-        if (errno == ERROR_NO_DATA) //TODO: pipe disconnected error?
-        {
-            std::cerr << "WARNING: Client disconnected, the rest of the output will be discarded" << endl;
-            inf->clientDisconnected = true;
-        }
-        return;
-    }
-
-    int outCode = MCMD_PARTIALOUT;
-    DWORD n;
-    if (!WriteFile(outNamedPipe,(const char*)&outCode, sizeof(outCode), &n, NULL))
-    {
-        LOG_err << "ERROR writing output Code to namedPipe: " << ERRNO;
-        if (errno == ERROR_NO_DATA) //TODO: pipe disconnected error?
+        if (ERRNO == ERROR_NO_DATA)
         {
             std::cerr << "WARNING: Client disconnected, the rest of the output will be discarded" << endl;
             inf->clientDisconnected = true;
@@ -323,25 +316,46 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, OUTSTRI
     }
 
     string sutf8;
-    localwtostring(s,&sutf8);
+    localwtostring(s, &sutf8);
+    assert(isValidUtf8(sutf8)); // above localwtostring should produce an empty string if the input had broken encoding or valid utf-8 otherwise
+
+    int outCode = MCMD_PARTIALOUT;
+    DWORD n;
+    if (!WriteFile(outNamedPipe,(const char*)&outCode, sizeof(outCode), &n, NULL))
+    {
+        std::cerr << "ERROR writing output Code to namedPipe: " << ERRNO << std::endl;
+        if (ERRNO == ERROR_NO_DATA) //TODO: pipe disconnected error?
+        {
+            std::cerr << "WARNING: Client disconnected, the rest of the output will be discarded" << endl;
+            inf->clientDisconnected = true;
+        }
+        return;
+    }
 
     size_t size = sutf8.size() > 1 ? sutf8.size() : 1; // client does not like empty responses
     if (!WriteFile(outNamedPipe,(const char*)&size, sizeof(size), &n, NULL))
     {
-        LOG_err << "ERROR writing output Code to namedPipe: " << ERRNO;
+        std::cerr << "ERROR writing output Code to namedPipe: " << std::endl;
         return;
     }
     if (!WriteFile(outNamedPipe,sutf8.data(), DWORD(size), &n, NULL))
     {
-        LOG_err << "ERROR writing to namedPipe: " << ERRNO;
+        std::cerr << "ERROR writing to namedPipe: " << ERRNO << std::endl;
     }
 }
 
 
-void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, char *s, size_t size)
+void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, char *s, size_t size, bool binaryContents)
 {
     HANDLE outNamedPipe = ((CmdPetitionNamedPipes *)inf)->outNamedPipe;
 
+    if (!binaryContents &&!isValidUtf8(s, size))
+    {
+        std::cerr << "Attempt to sendPartialOutput of invalid utf8 of size " << size << std::endl;
+        assert(false && "Attempt to sendPartialOutput of invalid utf8");
+        return;
+    }
+
     bool connectsucceeded = false;
     int attempts = 10;
     while (--attempts && !connectsucceeded)
@@ -354,6 +368,10 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, char *s
                 connectsucceeded = true;
                 break;
             }
+            else if (ERRNO == ERROR_NO_DATA)
+            {
+                break;
+            }
             else
             {
                 cerr << "ERROR on connecting to namedPipe " << outNamedPipe << ". errno: " << ERRNO << ". Attempts: " << attempts << endl;
@@ -369,7 +387,7 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, char *s
     if (!connectsucceeded)
     {
         cerr << "sendPartialOutput: Unable to connect on outnamedPipe " << outNamedPipe << " error: " << ERRNO << endl;
-        if (errno == ERROR_NO_DATA)
+        if (ERRNO == ERROR_NO_DATA)
         {
             std::cerr << "WARNING: Client disconnected, the rest of the output will be discarded" << endl;
             inf->clientDisconnected = true;
@@ -382,7 +400,7 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, char *s
     if (!WriteFile(outNamedPipe,(const char*)&outCode, sizeof(outCode), &n, NULL))
     {
         LOG_err << "ERROR writing output Code to namedPipe: " << ERRNO;
-        if (errno == ERROR_NO_DATA)
+        if (ERRNO == ERROR_NO_DATA)
         {
             std::cerr << "WARNING: Client disconnected, the rest of the output will be discarded" << endl;
             inf->clientDisconnected = true;
@@ -402,8 +420,15 @@ void ComunicationsManagerNamedPipes::sendPartialOutput(CmdPetition *inf, char *s
     }
 }
 
-int ComunicationsManagerNamedPipes::informStateListener(CmdPetition *inf, string &s)
+int ComunicationsManagerNamedPipes::informStateListener(CmdPetition *inf, const string &s)
 {
+    if (!isValidUtf8(s))
+    {
+        LOG_err << "Attempt to write an invalid utf8 string of size " << s.size();
+        assert(false && "Attempt to write an invalid utf8 string");
+        return 0;
+    }
+
     std::lock_guard<std::mutex> g(*informerMutex);
 
     LOG_verbose << "Inform State Listener: Output to write in namedPipe " << ((CmdPetitionNamedPipes *)inf)->outNamedPipe << ": <<" << s << ">>";
@@ -448,9 +473,9 @@ int ComunicationsManagerNamedPipes::informStateListener(CmdPetition *inf, string
  * @brief getPetition
  * @return pointer to new CmdPetitionPosix. Petition returned must be properly deleted (this can be calling returnAndClosePetition)
  */
-CmdPetition * ComunicationsManagerNamedPipes::getPetition()
+std::unique_ptr<CmdPetition> ComunicationsManagerNamedPipes::getPetition()
 {
-    CmdPetitionNamedPipes *inf = new CmdPetitionNamedPipes();
+    auto inf = std::make_unique<CmdPetitionNamedPipes>();
 
     wstring wread;
     wchar_t wbuffer[1024]= {};
@@ -484,7 +509,7 @@ CmdPetition * ComunicationsManagerNamedPipes::getPetition()
     if (!readok)
     {
         LOG_err << "Failed to read petition from named pipe. errno: L" << ERRNO;
-        inf->line = strdup("ERROR");
+        inf->line = "ERROR";
         return inf;
     }
 
@@ -497,14 +522,14 @@ CmdPetition * ComunicationsManagerNamedPipes::getPetition()
     if (!namedPipeValid(inf->outNamedPipe) || !namedPipe_id)
     {
         LOG_fatal << "ERROR creating output namedPipe at getPetition";
-        inf->line = strdup("ERROR");
+        inf->line = "ERROR";
         return inf;
     }
 
     if(!WriteFile(pipeGeneral,(const char*)&namedPipe_id, sizeof( namedPipe_id ), &n, NULL))
     {
         LOG_fatal << "ERROR writing to namedPipe at getPetition: ERRNO = " << ERRNO;
-        inf->line = strdup("ERROR");
+        inf->line = "ERROR";
         return inf;
     }
 
@@ -513,7 +538,7 @@ CmdPetition * ComunicationsManagerNamedPipes::getPetition()
         LOG_fatal << " Error disconnecting from general pip. errno: " << ERRNO;
     }
 
-    inf->line = strdup(receivedutf8.c_str());
+    inf->line = receivedutf8;
 
     return inf;
 }
