@@ -43,100 +43,16 @@
 #include <iterator>
 
 #include <regex> //split
-
-#ifdef _WIN32
-namespace mega {
-//override for the log. This is required for compiling, otherwise SimpleLog won't compile.
-std::ostringstream & operator<< ( std::ostringstream & ostr, const std::wstring & str)
-{
-    std::string s;
-    megacmd::localwtostring(&str,&s);
-    ostr << s;
-    return ( ostr );
-}
-}
-#endif
+#include <random>
 
 namespace megacmd {
 using namespace std;
 
-#ifdef _WIN32
-std::wostream & operator<< ( std::wostream & ostr, std::string const & str )
+std::string errorCodeStr(const std::error_code& ec)
 {
-    std::wstring toout;
-    stringtolocalw(str.c_str(),&toout);
-    ostr << toout;
-    return ( ostr );
+    return ec ? "(error: " + ec.message() + ")" : "";
 }
 
-std::wostream & operator<< ( std::wostream & ostr, const char * str )
-{
-    std::wstring toout;
-    stringtolocalw(str,&toout);
-    ostr << toout;
-    return ( ostr );
-}
-
-// convert UTF-8 to Windows Unicode wstring
-void stringtolocalw(const char* path, std::wstring* local)
-{
-    // make space for the worst case
-    local->resize((strlen(path) + 1) * sizeof(wchar_t));
-
-    int wchars_num = MultiByteToWideChar(CP_UTF8, 0, path,-1, NULL,0);
-    local->resize(wchars_num);
-
-    int len = MultiByteToWideChar(CP_UTF8, 0, path,-1, (wchar_t*)local->data(), wchars_num);
-
-    if (len)
-    {
-        local->resize(len-1);
-    }
-    else
-    {
-        local->clear();
-    }
-}
-
-//widechar to utf8 string
-void localwtostring(const std::wstring* wide, std::string *multibyte)
-{
-    if( !wide->empty() )
-    {
-        int size_needed = WideCharToMultiByte(CP_UTF8, 0, wide->data(), (int)wide->size(), NULL, 0, NULL, NULL);
-        multibyte->resize(size_needed);
-        WideCharToMultiByte(CP_UTF8, 0, wide->data(), (int)wide->size(), (char*)multibyte->data(), size_needed, NULL, NULL);
-    }
-}
-
-// convert Windows Unicode to UTF-8
-void utf16ToUtf8(const wchar_t* utf16data, int utf16size, string* utf8string)
-{
-    if(!utf16size)
-    {
-        utf8string->clear();
-        return;
-    }
-
-    utf8string->resize((utf16size + 1) * 4);
-
-    utf8string->resize(WideCharToMultiByte(CP_UTF8, 0, utf16data,
-        utf16size,
-        (char*)utf8string->data(),
-        int(utf8string->size() + 1),
-        NULL, NULL));
-}
-
-std::string getutf8fromUtf16(const wchar_t *ws)
-{
-    string utf8s;
-    utf16ToUtf8(ws, int(wcslen(ws)), &utf8s);
-    return utf8s;
-}
-
-#endif
-
-
 bool canWrite(string path)
 {
 #ifdef _WIN32
@@ -377,6 +293,26 @@ bool hasWildCards(string &what)
     return what.find('*') != string::npos || what.find('?') != string::npos;
 }
 
+
+string generateRandomAlphaNumericString(size_t len)
+{
+    static const std::string alphabet =
+        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+        "abcdefghijklmnopqrstuvwxyz"
+        "0123456789";
+
+    thread_local static std::mt19937 generator(std::random_device{}());
+    thread_local static std::uniform_int_distribution<> distribution(0, alphabet.size() - 1);
+
+    std::string randomString;
+    randomString.reserve(len);
+    for (size_t i = 0; i < len; ++i)
+    {
+        randomString += alphabet[distribution(generator)];
+    }
+    return randomString;
+}
+
 std::vector<std::string> split(const std::string& input, const std::string& pattern)
 {
     size_t start = 0, end;
@@ -424,6 +360,12 @@ std::string &ltrim(std::string &s, const char &c)
     return s;
 }
 
+std::string_view ltrim(const std::string_view s, const char &c)
+{
+    size_t pos = s.find_first_not_of(c);
+    return s.substr(pos == std::string::npos ? s.size() : pos, s.length());
+}
+
 std::string &rtrim(std::string &s, const char &c)
 {
     size_t pos = s.find_last_of(c);
@@ -445,11 +387,11 @@ string removeTrailingSeparators(string &path)
     return rtrim(rtrim(path,'/'),'\\');
 }
 
-vector<string> getlistOfWords(char *ptr, bool escapeBackSlashInCompletion, bool ignoreTrailingSpaces)
+vector<string> getlistOfWords(const char *ptr, bool escapeBackSlashInCompletion, bool ignoreTrailingSpaces)
 {
     vector<string> words;
 
-    char* wptr;
+    const char* wptr = "";
 
     // split line into words with quoting and escaping
     for (;; )
@@ -522,16 +464,16 @@ vector<string> getlistOfWords(char *ptr, bool escapeBackSlashInCompletion, bool
 
             wptr = ptr;
 
-            char *prev = ptr;
+            const char *prev = ptr;
             //while ((unsigned char)*ptr > ' ')
             while ((*ptr != '\0') && !(*ptr ==' ' && *prev !='\\'))
             {
-                if (*ptr == '"') // if quote is found, look for the ending quote
+                if (*ptr == '"' && *(ptr+1) != '\0') // if quote is found, look for the ending quote
                 {
-                    while (*(ptr + 1) != '"' && *(ptr + 1))
+                    do
                     {
-                        ptr++;
-                    }
+                        ++ptr;
+                    } while (*ptr != '"' && *(ptr+1) != '\0');
                 }
                 prev = ptr;
                 ptr++;
@@ -708,7 +650,23 @@ string getRightAlignedString(const string origin, unsigned int minsize)
     return os.str();
 }
 
-void printCenteredLine(OUTSTREAMTYPE &os, string msj, unsigned int width, bool encapsulated)
+bool startsWith(const std::string_view str, const std::string_view prefix)
+{
+    return str.rfind(prefix, 0) == 0;
+}
+
+string toLower(const std::string& str)
+{
+    std::string lower = str;
+    for (char& c : lower)
+    {
+        c = std::tolower(c);
+    }
+    return lower;
+}
+
+template <typename Stream_T>
+void printCenteredLine(Stream_T &os, string msj, unsigned int width, bool encapsulated)
 {
     unsigned int msjsize = getstringutf8size(msj);
     bool overflowed = false;
@@ -729,6 +687,10 @@ void printCenteredLine(OUTSTREAMTYPE &os, string msj, unsigned int width, bool e
     os << endl;
 }
 
+// Instantiate those that can be used:
+template void printCenteredLine<OUTSTREAMTYPE>(OUTSTREAMTYPE &os, string msj, unsigned int width, bool encapsulated);
+template void printCenteredLine<std::ostringstream>(std::ostringstream &os, string msj, unsigned int width, bool encapsulated);
+
 void printCenteredContents(OUTSTREAMTYPE &os, string msj, unsigned int width, bool encapsulated)
 {
      string headfoot = " ";
@@ -817,6 +779,9 @@ void printCenteredLine(string msj, unsigned int width, bool encapsulated)
 {
     OUTSTRINGSTREAM os;
     printCenteredLine(os, msj, width, encapsulated);
+#ifdef _WIN32
+    WindowsUtf8StdoutGuard utf8Guard;
+#endif
     COUT << os.str();
 }
 
@@ -824,6 +789,12 @@ void printCenteredContents(string msj, unsigned int width, bool encapsulated)
 {
     OUTSTRINGSTREAM os;
     printCenteredContents(os, msj, width, encapsulated);
+#ifdef _WIN32
+    WindowsUtf8StdoutGuard utf8Guard;
+#else
+    StdoutMutexGuard stdoutGuard;
+#endif
+
     COUT << os.str();
 }
 
@@ -831,6 +802,11 @@ void printCenteredContentsCerr(string msj, unsigned int width, bool encapsulated
 {
     OUTSTRINGSTREAM os;
     printCenteredContents(os, msj, width, encapsulated);
+#ifdef _WIN32
+    WindowsUtf8StdoutGuard utf8Guard;
+#else
+    StdoutMutexGuard stdoutGuard;
+#endif
     CERR << os.str();
 }
 
@@ -870,6 +846,13 @@ void printPercentageLineCerr(const char *title, long long completed, long long t
         *ptr++ = '#';
     }
 
+#ifdef _WIN32
+    WindowsUtf8StdoutGuard utf8Guard;
+#else
+    StdoutMutexGuard stdoutGuard;
+#endif
+
+     // TODO: should this be CERR ? (`\r`) may not behave as wanted
     if (cleanLineAfter)
     {
         cerr << outputString << '\r' << flush;
@@ -890,13 +873,13 @@ string getOption(map<string, string> *cloptions, const char * optname, string de
     return cloptions->count(optname) ? ( *cloptions )[optname] : defaultValue;
 }
 
-std::pair<string, bool> getOptionOrFalse(const map<string, string>& cloptions, const char * optname)
+std::optional<string> getOptionAsOptional(const map<string, string>& cloptions, const char * optname)
 {
     if (cloptions.find(optname) == cloptions.end())
     {
-        return {"", false};
+        return {};
     }
-    return {cloptions.at(optname), true};
+    return cloptions.at(optname);
 }
 
 int getintOption(map<string, string> *cloptions, const char * optname, int defaultValue)
@@ -1008,7 +991,7 @@ string sizeToText(long long totalSize, bool equalizeUnitsLength, bool humanreada
             unit = "KB";
         }
         os << fixed << reducedSize;
-        os << " " << unit;
+        os << " " << (reducedSize == 0.0 ? " " : "") << unit;
     }
     else
     {
@@ -1201,9 +1184,9 @@ string &trimProperty(string &what)
     return what;
 }
 
-string getPropertyFromFile(const char *configFile, const char *propertyName)
+string getPropertyFromFile(const fs::path &configFilePath, const char *propertyName)
 {
-    ifstream infile(configFile);
+    ifstream infile(configFilePath);
     string line;
 
     while (getline(infile, line))
@@ -1280,6 +1263,13 @@ ColumnDisplayer::ColumnDisplayer(std::map<std::string, int> *clflags, std::map<s
 
 }
 
+OUTSTRING ColumnDisplayer::str(bool printHeader)
+{
+    OUTSTRINGSTREAM oss;
+    print(oss, printHeader);
+    return oss.str();
+}
+
 void ColumnDisplayer::printHeaders(OUTSTREAMTYPE &os)
 {
     print(os, getintOption(mCloptions, "client-width", getNumberOfCols(75)), true, true);
@@ -1460,6 +1450,11 @@ void ColumnDisplayer::print(OUTSTREAMTYPE &os, int fullWidth, bool printHeader,
     }
 }
 
+void ColumnDisplayer::clear()
+{
+    *this = ColumnDisplayer(mClflags, mCloptions);
+}
+
 Field::Field()
 {
 
@@ -1485,42 +1480,34 @@ void Field::updateMaxValue(int newcandidate)
 std::unique_ptr<PlatformDirectories> PlatformDirectories::getPlatformSpecificDirectories()
 {
 #ifdef _WIN32
-    return std::unique_ptr<PlatformDirectories>(new WindowsDirectories);
+    return std::make_unique<WindowsDirectories>();
 #elif defined(__APPLE__)
-    return std::unique_ptr<PlatformDirectories>(new MacOSDirectories);
+    return std::make_unique<MacOSDirectories>();
 #else
-    return std::unique_ptr<PlatformDirectories>(new PosixDirectories);
+    return std::make_unique<PosixDirectories>();
 #endif
 }
 
 #ifdef _WIN32
-std::string WindowsDirectories::configDirPath()
+fs::path WindowsDirectories::configDirPath()
 {
-    TCHAR szPath[MAX_PATH];
-    std::string folder;
+    wchar_t szPath[MAX_PATH];
 
-    if (!SUCCEEDED(GetModuleFileName(NULL, szPath, MAX_PATH)))
+    if (GetModuleFileNameW(NULL, szPath, MAX_PATH) == 0)
     {
-        return std::string();
-    }
-    else
-    {
-        if (SUCCEEDED(PathRemoveFileSpec(szPath)))
-        {
-            if (PathAppend(szPath, TEXT(".megaCmd")))
-            {
-                utf16ToUtf8(szPath, lstrlen(szPath), &folder);
-            }
-        }
+        return {};
     }
 
-    auto suffix = getenv("MEGACMD_WORKING_FOLDER_SUFFIX");
+    fs::path folder = fs::path(szPath).parent_path() / ".megaCmd";
+
+    auto suffix = _wgetenv(L"MEGACMD_WORKING_FOLDER_SUFFIX");
     if (suffix != nullptr)
     {
         folder += "_";
         folder += suffix;
     }
 
+    folder = nonMaxPathLimitedPath(folder); // prepend //?/
     return folder;
 }
 
@@ -1565,7 +1552,7 @@ std::string PosixDirectories::homeDirPath()
     return std::string(pwd.pw_dir);
 }
 
-std::string PosixDirectories::configDirPath()
+fs::path PosixDirectories::configDirPath()
 {
     std::string home = homeDirPath();
     if (home.empty())
@@ -1585,7 +1572,7 @@ string PosixDirectories::noHomeFallbackFolder()
 }
 
 #ifdef __APPLE__
-std::string MacOSDirectories::runtimeDirPath()
+fs::path MacOSDirectories::runtimeDirPath()
 {
     std::string home = homeDirPath();
     if (home.empty())
@@ -1617,9 +1604,12 @@ std::string getOrCreateSocketPath(bool createDirectory)
 
     static auto MAX_SOCKET_PATH = sizeof(sockaddr_un::sun_path) / sizeof(decltype(sockaddr_un::sun_path[0]));
 
-    if ((socketFolder.size() + 1 + sockname.size()) >= (MAX_SOCKET_PATH - 1))
+    if ((socketFolder.string().size() + 1 + sockname.size()) >= (MAX_SOCKET_PATH - 1))
     {
-        std::cerr << "WARN: socket path in runtime dir would exceed max size. Falling back to /tmp" << std::endl;
+        if (createDirectory)
+        {
+            std::cerr << "WARN: socket path in runtime dir would exceed max size. Falling back to /tmp" << std::endl;
+        }
         socketFolder = PosixDirectories::noHomeFallbackFolder();
     }
 
@@ -1638,11 +1628,13 @@ std::string getOrCreateSocketPath(bool createDirectory)
             umask(mode);
 
             if (failed)
+            {
                 return std::string();
+            }
         }
     }
 
-    return socketFolder.append("/").append(sockname);
+    return (socketFolder / sockname).string();
 }
 #endif // ifdef(_WIN32) else
 } //end namespace
